from dataclasses import dataclass, field
from multiprocessing import Process
from multiprocessing.connection import Connection
from threading import Thread
from typing import Type

from edri.dataclass.event import Event


@dataclass
class Worker:
    pipe: Connection
    event: Event | None
    worker: Thread | Process
    streams: dict[Type[Event], str] = field(default_factory=dict)



from json import loads, JSONDecodeError, dumps
from typing import Callable, Type, Unpack, TypedDict

from edri.api import Headers
from edri.api.dataclass.api_event import api_events
from edri.api.handlers import BaseHandler
from edri.config.constant import ApiType
from edri.dataclass.directive import ResponseDirective
from edri.dataclass.event import Event
from edri.dataclass.response import ResponseStatus
from edri.utility import NormalizedDefaultDict
from edri.utility.function import camel2snake


class ResponseKW(TypedDict):
    headers: NormalizedDefaultDict[str, Headers]


class WebsocketHandler(BaseHandler):
    _events = None
    _commands = None

    def handle_directives(self, directives: list[ResponseDirective]) -> ...:
        pass

    def __init__(self, scope: dict, receive: Callable, send: Callable, headers: NormalizedDefaultDict[str, Headers]):
        super().__init__(scope, receive, send)
        self.command: str | None = None
        self.headers = headers

    @classmethod
    def events(cls) -> dict[str, Type[Event]]:
        if cls._events is None:
            cls._events, cls._commands = cls.sort_events()
        return cls._events

    @classmethod
    def commands(cls) -> dict[Type[Event], str]:
        if cls._commands is None:
            cls._events, cls._commands = cls.sort_events()
        return cls._commands

    @staticmethod
    def sort_events() -> tuple[dict[str, Type[Event]], dict[Type[Event], str]]:
        resources = {}
        for event in api_events:
            if ApiType.WS in event.exclude:
                continue
            if event.resource in resources:
                raise KeyError(f"Duplicate key found: {event.resource}")
            resources[event.resource] = event.event

        return resources, {event.event: event.resource for event in api_events if ApiType.WS not in event.exclude}

    @classmethod
    def type(cls) -> ApiType:
        return ApiType.WS

    async def accept_client(self) -> bool:
        data = await self.receive()
        if "type" in data and data["type"] == "websocket.connect":
            await self.send({"type": "websocket.accept"})
            return True
        else:
            return False

    async def parse_body(self, data) -> bool:
        if data["type"] == "websocket.receive":
            if data["text"] is not None:
                received_data = data["text"]
            else:
                received_data = data["bytes"].decode("utf-8", errors="replace")
        elif data["type"] == "websocket.disconnect":
            return False
        else:
            self.logger.error("Parse body failed")
            await self.response_error(1003, {
                "status": ResponseStatus.FAILED.name,
                "reasons": ["Parse body failed"]
            })
            return False
        try:
            self.parameters.update({camel2snake(key): value for key, value in loads(received_data).items()})
        except JSONDecodeError as e:
            self.logger.warning("Cannot process json data", exc_info=e)
            await self.response_error(1003, {
                "status": ResponseStatus.FAILED.name,
                "reasons": ["Cannot process json data"]
            })
            return False
        except Exception as e:
            self.logger.error("Unknown error", exc_info=e)
            await self.response_error(1002, {
                "status": ResponseStatus.FAILED.name,
                "reasons": ["Unknown error"]
            })
            return False
        return True

    async def get_event_constructor(self) -> Type[Event] | None:
        self.command = self.parameters.pop("command", None)
        if self.command is None:
            raise ResourceWarning("Missing command")
        return self.events().get(self.command, None)

    async def response(self, status, data, *args, **kwargs: Unpack[ResponseKW]) -> None:
        event = {
            "command": self.commands()[data.__class__],
        }
        event.update(data.as_dict(transform=True, keep_concealed=False))
        response = data.get_response()
        if response:
            event["response"] = response.as_dict(transform=True, keep_concealed=False)
        else:
            event.pop("response")
        await self.send({"type": "websocket.send", "text": dumps(event)})

    async def response_error(self, status, data: dict | None = None, *args, **kwargs) -> None:
        if status:
            await self.send({"type": "websocket.close", "code": status, "reason": dumps(data)})
        elif isinstance(data, dict):
            await self.send({"type": "websocket.send", "text": dumps(data)})

    async def clear(self):
        self.headers = {}
        self.parameters = {}
        self.command = None



from collections import defaultdict
from logging import getLogger
from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
from multiprocessing.queues import Queue as MPQueue
from threading import Thread
from time import sleep
from typing import Type
from uuid import UUID, uuid1

from edri.config.setting import CACHE_TIMEOUT
from edri.dataclass.event import Event, SwitchInfo
from edri.events.edri.router import SubscribeConnector, SubscribedExternal, SubscribedNew, SendFrom, LastEvents, Demands
from edri.router.connector import Socket
from edri.utility.queue import Queue


class Connector(Process):
    def __init__(self, router_queue: MPQueue[Event], router_id: UUID | None = None) -> None:
        super(Connector, self).__init__()
        self.router_queue = router_queue
        self.local_requests: set[Type[Event]] = set()
        self.local_responses: set[Type[Event]] = set()
        self.remote_request: dict[Type[Event], set[UUID]] = defaultdict(set)
        self.remote_response: dict[Type[Event], set[UUID]] = defaultdict(set)
        self.logger = getLogger(__name__)
        self.router_pipe: Connection | None = None

        self.sender_queue: Queue[tuple[Event, UUID | None]] | None = None
        self.sender_thread: Thread | None = None
        self.receiver_thread: Thread | None = None

        if router_id is not None:
            self.router_id = router_id
        else:
            self.router_id = uuid1(clock_seq=0)
        self.socket = Socket()
        if CACHE_TIMEOUT < 10:
            raise ValueError("CACHE_TIMEOUT must be greater than 10s")

    def sender(self) -> None:
        while True:
            event, router_id = self.sender_queue.get()
            if event is False:
                break
            event._switch = SwitchInfo(router_id=self.router_id)
            self.logger.debug("Sender sending event type: %s", event.__class__)
            if router_id:
                try:
                    self.socket.send_router(router_id, event)
                except BrokenPipeError as e:
                    self.logger.warning("Event cannot be send: %s", event, exc_info=e)
                    self.sender_queue.unget((event, router_id))
                    self.socket.reconnect()
                    return
            if not event.has_response():
                router_id_list = self.remote_request[event.__class__]
            else:
                router_id_list = self.remote_response[event.__class__]
            for router_id in router_id_list:
                try:
                    self.socket.send_router(router_id, event)
                except BrokenPipeError as e:
                    self.logger.warning("Event cannot be send: %s", event, exc_info=e)
                    self.sender_queue.unget((event, None))
                    self.socket.reconnect()
                    return

        self.logger.error("Sender se ukončil")

    def receiver(self) -> None:
        while True:
            event = self.socket.receive()
            self.logger.debug("New event: %s", event)
            if isinstance(event, Demands):
                self.logger.info("New event with demands from remote router: %s", event.router_id)
                self.logger.debug("Demands - requests: %s responses: %s", event.requests, event.responses)
                for demand in event.requests:
                    self.remote_request[demand].add(event.router_id)
                    subscribe = SubscribeConnector(event=demand, request=True)
                    self.router_queue.put(subscribe)

                for demand in event.responses:
                    self.remote_response[demand].add(event.router_id)
                    subscribe = SubscribeConnector(event=demand, request=False)
                    self.router_queue.put(subscribe)
            elif isinstance(event, LastEvents):
                self.logger.debug("Request for cached events from remote router: %s", event.router_id)
                last_event = event.response.last_events.get(self.router_id, None)
                if last_event:
                    self.router_queue.put(SendFrom(router_id=event.router_id, key=last_event))
                else:
                    self.router_queue.put(SendFrom(router_id=event.router_id))
            else:
                if event._switch is None:
                    self.logger.warning("Event has no switch information: %s", event)
                    continue
                event._switch.received = True
                self.logger.info("-> Router %s", event)
                self.router_queue.put(event)


    def request_last_events(self) -> None:
        self.router_queue.put(LastEvents(self.router_id))

    def send_basic_info(self) -> None:
        self.logger.debug("Sending information to switch")
        self.socket.send_router_id(self.router_id)
        self.socket.send_demands(self.router_id, self.local_requests, self.local_responses)
        self.request_last_events()

    def run(self) -> None:
        sleep(5)
        self.logger.debug(f"SwitchConnector is running!")
        self.logger.info("Router ID: %s", self.router_id)
        self.socket.reconnect()
        self.router_pipe, pipe = Pipe(duplex=False)
        subscribe_connector = SubscribedExternal(pipe=pipe)
        self.router_queue.put(subscribe_connector)
        while True:
            event = self.router_pipe.recv()
            if isinstance(event, SubscribedExternal):
                self.logger.debug("Received %s events from router", len(event.response.demands.requests) + len(event.response.demands.responses))
                self.local_requests = event.response.demands.requests
                self.local_responses = event.response.demands.responses
                break
            else:
                self.logger.error("Unexpected event: %s", event)
        pipe.close()

        self.send_basic_info()

        self.receiver_thread = Thread(target=self.receiver)
        self.receiver_thread.start()

        self.sender_queue = Queue()
        self.sender_thread = Thread(target=self.sender)
        self.sender_thread.start()

        while True:
            if self.router_pipe.poll(timeout=5):
                event = self.router_pipe.recv()
                self.logger.debug("<- Router - %s", event)
                if isinstance(event, SubscribedNew):
                    if event.request:
                        self.local_requests.add(event.event)
                    else:
                        self.local_responses.add(event.event)
                    self.socket.send_demands(self.router_id, self.local_requests, self.local_responses)
                elif isinstance(event, LastEvents):
                    self.socket.send_request_last_events(event)
                elif isinstance(event, SendFrom):
                    self.sender_queue.put((event.response.event, event.router_id))
                else:
                    self.sender_queue.put((event, None))
            if not self.receiver_thread.is_alive():
                self.logger.error("Receiver was shut down")
                self.sender_queue.put((False, None))
                while self.sender_thread.is_alive():
                    sleep(1)
                self.socket.reconnect()
                self.send_basic_info()
                self.receiver_thread = Thread(target=self.receiver)
                self.receiver_thread.start()

                self.sender_thread = Thread(target=self.sender)
                self.sender_thread.start()

    def quit(self) -> None:
        if self.sender_queue:
            self.sender_queue.put((False, None))
        self.socket.close()



from abc import ABC, ABCMeta
from dataclasses import dataclass
from enum import Enum, auto
from typing import Self, Optional, TYPE_CHECKING

from edri.dataclass.event import Event

if TYPE_CHECKING:
    from edri.api.broker import EventInfo


class MiddlewareMeta(ABCMeta):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)

        # Dynamically create the list of methods from the class being created
        required_methods = []
        for base in cls.__mro__:
            for method_name, method in base.__dict__.items():
                if callable(method) and not method_name.startswith('__'):
                    required_methods.append(method_name)

        # Check if at least one method is implemented in the subclass
        if not any(attr in cls.__dict__ for attr in required_methods):
            raise TypeError(
                f"Class '{cls.__name__}' must implement at least one of the following methods: {", ".join(required_methods)}. "
                "Ensure that your class provides an implementation for at least one of these methods."
            )


@dataclass
class MiddlewareControl:
    """
    Encapsulates the outcome of middleware processing with details on the action type, optional data, errors,
    and an optional replacement event.
    """

    class ActionType(Enum):
        """Enum to define middleware processing actions."""
        CONTINUE = auto()
        STOP = auto()
        REPLACE_EVENT = auto()

    action: ActionType
    event: Optional[Event] = None

    @classmethod
    def continue_processing(cls) -> Self:
        """Creates a MiddlewareControl instance indicating processing should continue."""
        return cls(action=cls.ActionType.CONTINUE)

    @classmethod
    def stop_processing(cls) -> Self:
        """Creates a MiddlewareControl instance indicating processing should stop."""
        return cls(action=cls.ActionType.STOP)

    @classmethod
    def replace_event(cls, event: Event) -> Self:
        """
        Creates a MiddlewareControl instance that replaces the current event with a new one.

        Args:
            event (Event): The event that should replace the current event in processing.

        Returns:
            MiddlewareControl: An instance with action set to REPLACE_EVENT.
        """
        return cls(action=cls.ActionType.REPLACE_EVENT, event=event)


class Middleware(ABC, metaclass=MiddlewareMeta):
    """
    Abstract base class for middleware components.
    Middleware classes derived from this must implement at least one of
    'process_request' and 'process_response'.
    """

    def process_request(self, event: Event, info: "EventInfo") -> MiddlewareControl:
        """
        Method to process the request before it reaches the main application logic.
        This method should be implemented in a subclass to perform any necessary
        preprocessing or validation on the request.

        :param event: The request event to be processed.
        :param info: Information about the event.
        :return: A `MiddlewareControl` object indicating whether to continue processing,
                 skip further middleware, or terminate the request/response cycle.
        """
        raise NotImplementedError("Must implement process_request or process_response")

    def process_response(self, event: Event, info: "EventInfo") -> MiddlewareControl:
        """
        Method to process the response before it is sent to the client.
        This method should be implemented in a subclass to perform any necessary
        postprocessing or formatting on the response.

        :param event: The response event to be processed.
        :param info: Information about the event.
        :return: A `MiddlewareControl` object indicating whether to continue processing,
                 skip further middleware, or terminate the request/response cycle.
        """
        raise NotImplementedError("Must implement process_request or process_response")

    @classmethod
    def __getattr__(cls, name):
        prefix = "is_"
        if name.startswith(prefix):
            capability_name = f"process_{name[len(prefix):]}"
            method = cls.__dict__.get(capability_name)
            return callable(method)
        raise AttributeError(f"type object '{cls.__name__}' has no attribute '{name}'")



from enum import IntEnum, unique, auto

SWITCH_BYTES_LENGTH = 16  # Number of bytes in the switch identifier
SWITCH_MAX_SIZE = 0xFFFFFFFF  # Maximum value for a 32-bit unsigned integer
SWITCH_COMMANDS = (1 << SWITCH_BYTES_LENGTH * 8) - 1  # Total number of possible commands
SCHEDULER_TIMEOUT_MAX = 2_147_483  # Maximum timeout value for the scheduler, equivalent to INT_MAX without the last three digits
STREAM_CLOSE_MARK = "#&@"  # Marker used to indicate the end of a stream


@unique
class SwitchMessages(IntEnum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return last_values[-1] - count

    NEW_DEMANDS = SWITCH_COMMANDS
    LAST_MESSAGES = auto()


@unique
class ApiType(IntEnum):
    WS = 0
    REST = auto()
    HTML = auto()



from typing import Callable, Optional, Any


class StringTransformer(str):
    """
    A class that extends str and allows transformation on initialization.
    The transformation can be predefined (like lowercasing) or custom using a lambda function.
    The transformation settings are applied during object initialization.

    Args:
        transform (callable, optional): A function that takes a string and returns a transformed version.
        lower (bool, optional): If True, the string will be converted to lowercase.
        upper (bool, optional): If True, the string will be converted to uppercase.

    Example:
        >>> st = StringTransformer("Hello", lower=True)
        >>> print(st)
        'hello'

        >>> st = StringTransformer("Hello", transform=lambda x: x[::-1])
        >>> print(st)
        'olleH'
    """

    def __new__(cls, value: Any,
                transform: Optional[Callable[[str], str]] = None,
                lower: bool = False,
                upper: bool = False):
        value = super().__new__(cls, value)
        # Apply predefined transformations (lower/upper)
        if lower:
            value = value.lower()
        if upper:
            value = value.upper()

        # Apply the custom transformation function if provided
        if transform is not None:
            value = transform(value)

        # Return the transformed value as an instance of StringTransformer
        return value


from dataclasses import is_dataclass, asdict
from enum import Enum
from json import JSONEncoder
from pathlib import Path
from typing import Any
from uuid import UUID


class CustomJSONEncoder(JSONEncoder):
    """
    Custom JSON encoder that extends Python's built-in JSONEncoder to handle additional Python types
    and custom serialization logic.

    Supported types and custom behaviors:
    - Objects with a `to_json()` method: Serialized using that method.
    - Datetime-like objects with `isoformat()`: Serialized to ISO 8601 format.
    - `Path` objects: Serialized as POSIX-style strings.
    - `bytes` and `bytearray`: Encoded as hexadecimal strings.
    - `Enum` members: Serialized using their `.value`.
    - `UUID` instances: Serialized as hexadecimal strings.
    - `Exception` instances: Serialized as a dictionary with type, message, and args.
    - Dataclass instances: Converted to dictionaries via `asdict`.

    Args:
        skipkeys (bool): Whether to skip keys not of a basic type (str, int, float, bool, None).
        ensure_ascii (bool): If True, escapes all non-ASCII characters in output.
        check_circular (bool): If True, checks for circular references and raises an error if found.
        allow_nan (bool): If True, allows NaN and Infinity values (JavaScript-compatible).
        sort_keys (bool): If True, output dictionary keys are sorted.
        indent (int or None): If specified, pretty-prints output with given number of spaces.
        separators (tuple or None): Tuple of (item_separator, key_separator) for customizing output.
        default (callable or None): A fallback function for unsupported objects.
        context (dict or None): Optional context dictionary for passing extra encoding parameters.
    """

    def __init__(self, *, skipkeys=False, ensure_ascii=True, check_circular=True, allow_nan=True, sort_keys=False,
                 indent=None, separators=None, default=None, context=None):
        super().__init__(
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            sort_keys=sort_keys,
            indent=indent,
            separators=separators,
            default=default,
        )
        self.context = context

    def default(self, data: Any) -> Any:
        """
        Serializes unsupported data types into JSON-compatible formats.

        Custom serialization behavior:
        - If object has `to_json()` method: Calls and returns its result.
        - If object has `isoformat()` method (e.g., datetime): Uses its ISO 8601 representation.
        - If object is a `Path`: Returns its POSIX-style path string.
        - If object is `bytes` or `bytearray`: Returns a hex-encoded string.
        - If object is an `Enum`: Returns its `.value`.
        - If object is a `UUID`: Returns its hexadecimal string.
        - If object is an `Exception`: Returns a dict with type, message, and args.
        - If object is a dataclass: Converts to a dictionary via `asdict()`.

        Args:
            data (Any): The object to be serialized.

        Returns:
            Any: A JSON-serializable object.
        """
        if hasattr(data, "to_json"):
            return data.to_json()
        elif hasattr(data, "isoformat"):
            return data.isoformat()
        elif isinstance(data, Path):
            return data.as_posix()
        elif isinstance(data, (bytes, bytearray)):
            return data.hex()
        elif isinstance(data, Enum):
            return data.value
        elif isinstance(data, UUID):
            return data.hex
        elif isinstance(data, Exception):
            return {
                'type': type(data).__name__,
                'message': str(data),
                'args': data.args
            }
        elif is_dataclass(data):
            return asdict(data)
        else:
            return super().default(data)



from collections import defaultdict
from dataclasses import dataclass, fields
from logging import getLogger
from multiprocessing.connection import wait, Connection
from multiprocessing.queues import Queue
from typing import Any, DefaultDict, Type, Never

from edri.abstract.manager.manager_base import ManagerBase
from edri.api.dataclass import Client
from edri.api.dataclass.api_event import api_events
from edri.api.middleware import Middleware, MiddlewareControl
from edri.config.constant import ApiType
from edri.dataclass.directive.base import InternalServerErrorResponseDirective
from edri.dataclass.event import Event, EventHandlingType, ApiInfo
from edri.dataclass.response import ResponseStatus
from edri.events.api import group, manage, client
# from edri.events.api.client.documentation import Documentation
from edri.events.edri.router import HealthCheck
from edri.utility import Storage


class TrieNode:
    def __init__(self, key=None):
        self.key = key
        self.children: dict[tuple[str, Any], TrieNode] = {}
        self.is_end: bool = False
        self.clients: set[Client] = set()

    def __repr__(self) -> str:
        """
        Returns a tree-like string representation of the trie starting from this node.

        Returns:
            str: String representation of the trie starting from this node.
        """
        return self._build_repr(level=0)

    def _build_repr(self, level: int = 0, max_depth: int | None = None, max_children: int | None = None) -> str:
        """
        Recursively builds the string representation of the trie.

        Parameters:
            level (int): Current depth level in the trie.
            max_depth (int | None): Maximum depth to traverse for the representation.
            max_children (int | None): Maximum number of children to display per node.

        Returns:
            str: String representation of the trie starting from this node.
        """
        indent = "    " * level
        node_label = f"TrieNode(key={self.key}, is_end={self.is_end}, clients={len(self.clients)})"
        representation = f"{indent}{node_label}\n"

        if max_depth is not None and level >= max_depth:
            return representation

        children_keys = sorted(self.children.keys())
        displayed_children = children_keys
        if max_children is not None and len(children_keys) > max_children:
            displayed_children = children_keys[:max_children] + ['...']

        for child_key in displayed_children:
            if child_key == '...':
                representation += f"{indent}    ...\n"
                break
            child_node = self.children[child_key]
            representation += child_node._build_repr(level=level + 1, max_depth=max_depth, max_children=max_children)

        return representation


@dataclass
class EventInfo:
    resource: str
    handling_type: EventHandlingType


class Broker(ManagerBase):
    """
    The ApiBroker class acts as an intermediary between the central router and all API handlers,
    managing client registrations, event dispatching, and communication in a multiprocessing
    environment. It extends ManagerBase to leverage common management functionalities and
    adds capabilities specific to API interactions.

    Attributes:
        api_broker_queue (Queue): A multiprocessing queue used to receive events specifically for the API broker.
        clients (Optional[Storage[Client]]): A storage mechanism for registered clients, allowing for quick access and management.
        db: Placeholder attribute for potential database connections or operations, not implemented by default.
        events (dict): A mapping of event resources to event classes, derived from the global api_events definition.
        middlewares_request (list[Middleware]): List of middlewares applied to incoming requests.
        middlewares_response (list[Middleware]): List of middlewares applied to outgoing responses.

    Methods:
        __init__(router_queue: "Queue[Event]", api_broker_queue: "Queue[Event]", middlewares: list[Middleware]):
            Constructor for the ApiBroker class.
        name (property): Returns the name of the ApiBroker instance.
        send_specific(event: Event): Sends an event to a specific client based on the event's key.
        send_all(event: Event): Broadcasts an event to all clients of a specific API type.
        client_register(event: client.Register): Registers a new client and sends a confirmation event.
        client_unregister(event: Event): Unregisters a client based on the event's key.
        event_register(event: manage.Register): Registers a client for specific events and parameters.
        solve(event: Event): Determines the appropriate method to handle an incoming event.
        _prepare(): Prepares the ApiBroker for operation by initializing storage for clients.
        additional_pipes (property): Returns a set of additional pipes to monitor for incoming events.
        run_resolver(): Main loop for processing incoming events and dispatching them to the appropriate handlers.
        resolve_callback(event: Event, pipe: Connection): Callback method for handling events received from clients.
        _prepare_resolvers(): Prepares resolvers for handling specific types of events.
        resolve(event: Event): Processes an event through the appropriate middlewares and forwards it to its destination.

    Usage:
        This class is instantiated with specific queues and is used within a multiprocessing
        application to manage communications between the router and API handlers. It handles
        client registrations, event dispatching, and utilizes middlewares for request and
        response processing.
    """

    def __init__(self, router_queue: Queue[Event], api_broker_queue: Queue[Event], middlewares: list[Middleware]) -> None:
        """
        Initializes an ApiBroker instance with specific queues for routing and
        API broker events.

        Parameters:
            router_queue (Queue): The queue for receiving routed events from other
            components of the system.
            api_broker_queue (Queue): The queue specifically for the API broker to
            receive events.

        Returns:
            None
        """
        super().__init__(router_queue)
        self.api_broker_queue = api_broker_queue
        self.logger = getLogger(__name__)
        self.clients: Storage[Client]
        self.db = None
        self.events = {event.event: EventInfo(event.resource, event.handling) for event in api_events}
        self.middlewares_request = [middleware for middleware in middlewares if middleware.is_request]
        self.middlewares_response = [middleware for middleware in middlewares if middleware.is_response]
        self.roots: DefaultDict[Type[Event], TrieNode] = defaultdict(TrieNode)

    @property
    def name(self) -> str:
        """
        Property returning the name of the ApiBroker instance. Primarily used for
        logging and identification purposes.

        Returns:
            str: The name "ApiBroker".
        """
        return "ApiBroker"

    def _send(self, client: Client, event: Event) -> None:
        """
        Attempts to send an event to a specific client. If the connection is broken
        or reset during sending, it logs a warning and unregisters the client.

        Parameters:
            client (Client): The client object representing the target for the event.
            event (Event): The event to be sent to the client.

        Returns:
            None
        """
        try:
            client.socket.send(event)
        except (BrokenPipeError, ConnectionResetError) as e:
            self.logger.warning("Cannot be send %s", client.socket, exc_info=e)
            self.client_unregister(event)

    def send_specific(self, event: Event) -> None:
        """
        Sends an event to the specific client identified by the event's key. Logs
        a warning if the client is not found.

        Parameters:
            event (Event): The event containing the key of the target client.

        Returns:
            None
        """
        self.logger.debug("-> %s", event)
        if not event._api or not event._api.key:
            self.logger.error("Key not found in the event %s", event)
            return
        try:
            client = self.clients[event._api.key]
        except KeyError:
            self.logger.warning("Client was not found: %s", event)
            return
        self._send(client, event)

    def send_subscribed(self, event: Event) -> None:
        """
        Sends an event to all clients that are subscribed to the parameters
        matching the event's attributes.

        Parameters:
            event (Event): The event to be sent to subscribed clients.

         Returns:
             None
        """

        clients = self._search_trie(self.roots[event.__class__], event)
        for client in clients:
            self.logger.debug("->? %s", event)
            self._send(client, event)

    def send_all(self, event: Event) -> None:
        """
         Broadcasts an event to all clients that are of a specific API type (e.g., WebSocket).
         It iterates through all clients and sends the event to those matching the type.

         Parameters:
             event (Event): The event to be broadcasted.

         Returns:
             None
         """
        self.logger.debug("->* %s", event)
        for client in self.clients.values():
            if client.type == ApiType.WS:
                self._send(client, event)

    def client_register(self, event: client.Register) -> None:
        """
        Handles the registration of a new client based on the received register event.
        It assigns a unique key to the client, stores it, and acknowledges the registration.

        Parameters:
            event (client.Register): The event containing the client's registration information.

        Returns:
            None
        """
        event._api = ApiInfo(self.clients.append(Client(event.socket, event.type)), event.type)
        event.socket = None
        event.response.set_status(ResponseStatus.OK)
        self.send_specific(event)

    def client_unregister(self, event: Event) -> None:
        """
        Handles the unregistration of a client, removing them from the stored clients
        based on the event's key. It performs cleanup by closing the client's socket
        and removing them from the trie.
        Parameters:
            event (Event): The event containing the key of the client to be unregistered.

        Returns:
            None
        """
        try:
            if event._api and event._api.key:
                client = self.clients[event._api.key]
                client.socket.close()
                del self.clients[event._api.key]
                self._remove_client_from_tries(client)
            else:
                self.logger.warning("Client cant be unregistered because key is missing!")
        except KeyError as e:
            self.logger.debug("Client was not found!", exc_info=e)

    def event_register(self, event: manage.Register) -> None:
        """
          Registers a client for specific events and their parameters, allowing for
          customized event handling. This method updates the client's event subscriptions
          and their associated parameters based on the provided event registration details.

          Parameters:
              event (manage.Register): An object representing the event registration request.
              This object must include a non-empty `_key` attribute identifying the client,
              a list of `events` the client wishes to subscribe to, and corresponding
              `parameters` and `values` for those events.

          Returns:
              None

          Raises:
              KeyError: If the event `_key` attribute is missing or if the `_key` does not
              correspond to any existing client in the `clients` dictionary.

              ValueError: If the `events` list in the `event` object is empty, indicating
              that no events have been specified for registration.

              LookupError: If any of the events specified in the `event.events` list do not
              exist in the `self.events` mapping, indicating an attempt to register for
              undefined events.

          This method first verifies that the event `_key` is present and corresponds to an
          existing client. It then proceeds to update the client's event subscriptions and
          parameters. If successful, the client's response status is set to `OK`, and the
          updated event registration details are sent back to the client via their socket.
          """
        if not event._api.key:
            raise AttributeError("Event key is missing!")
        try:
            client = self.clients[event._api.key]
        except KeyError as e:
            self.logger.error("Client was not found: %s", event, exc_info=e)
            event.response.set_status(ResponseStatus.FAILED)
            return

        try:
            root = self._find_root(event.event)
        except KeyError as e:
            self.logger.error("Event was not found: %s", event.event, exc_info=e)
            event.response.set_status(ResponseStatus.FAILED)
            client.socket.send(event)
            return

        self._insert_client_to_trie(root, event.param_set, client)

        event.response.set_status(ResponseStatus.OK)
        client.socket.send(event)

    def event_unregister(self, event: manage.Unregister) -> None:
        """
        Unregisters a client from a specific subscription based on the provided param_set.
        Removes the client from the trie along the path defined by param_set.

        Parameters:
            event (manage.Unregister): The event containing unregistration details, including param_set.

         Returns:
             None
        """
        if not event._api and not event._api.key:
            raise AttributeError("Event key is missing!")
        try:
            client = self.clients[event._api.key]
        except KeyError as e:
            self.logger.error("Client was not found: %s", event, exc_info=e)
            event.response.set_status(ResponseStatus.FAILED)
            return

        # Remove client from trie for the given param_set
        param_set = event.param_set
        param_items = list(param_set.items())
        try:
            root = self._find_root(event.event)
        except KeyError as e:
            self.logger.error("Event was not found: %s", event.event, exc_info=e)
            event.response.set_status(ResponseStatus.FAILED)
            client.socket.send(event)
            return

        removed = self._remove_client_from_trie_by_param_set(root, param_items, 0, client)

        if removed:
            self.logger.debug("Client %s successfully unregistered from param_set %s", client, param_set)
            event.response.set_status(ResponseStatus.OK)
        else:
            self.logger.warning("Client %s was not registered for param_set %s", client, param_set)
            event.response.set_status(ResponseStatus.FAILED)
        client.socket.send(event)

    # def documentation(self, event: Documentation):
    #     event.response.set_status(ResponseStatus.OK)
    #     self.send_specific(event)

    def solve(self, event: Event) -> None:
        """
        Determines the appropriate method to handle an incoming event based on its type and
        handling type specified in the events mapping.

        Parameters:
            event (Event): The event to be handled.

         Returns:
             None
        """
        if (event_info := self.events.get(event.__class__)) is None:
            self.logger.error("Unknown event type: %s", event.__class__)
        else:
            for middleware in self.middlewares_response:
                event._timing.stamp(middleware.__class__.__name__, "Started")
                control = middleware.process_response(event, event_info)
                event._timing.stamp(middleware.__class__.__name__, "Finished")
                if not isinstance(control, MiddlewareControl):
                    self.logger.error("Middleware %s suppose to return MiddlewareControl object", middleware.__class__.__name__)
                    event.response.set_status(ResponseStatus.FAILED)
                    event.response.add_directive(InternalServerErrorResponseDirective(
                        f"Middleware  {middleware.__class__.__name__} suppose to return MiddlewareControl object"))
                    self.resolve(event)
                    return

                if control.action == MiddlewareControl.ActionType.CONTINUE:
                    continue

                elif control.action == MiddlewareControl.ActionType.STOP:
                    self.logger.info("Processing stopped: %s", middleware.__class__.__name__)
                    break

                elif control.action == MiddlewareControl.ActionType.REPLACE_EVENT:
                    self.logger.error("Event cannot be replaced in response processing")


            handling_type = event_info.handling_type
            if handling_type & EventHandlingType.SPECIFIC and event._api and event._api.key:
                self.send_specific(event)
            elif handling_type & EventHandlingType.SUBSCRIBED:
                self.send_subscribed(event)
            elif handling_type & EventHandlingType.ALL:
                self.send_all(event)
            else:
                self.logger.warning("Event %s with unexpected handling type %s", event, handling_type)

    def _prepare(self) -> None:
        """
         Prepares the ApiBroker for operation by initializing necessary components,
         such as client storage.

         Returns:
             None
         """
        super()._prepare()
        self.clients = Storage[Client]()

    @property
    def additional_pipes(self) -> set[Connection]:
        """
        Provides a set of additional pipes that should be monitored for incoming
        events, specifically the API broker queue.

        Returns:
            Set[Connection]: A set containing the reader end of the API broker queue.
        """
        return {self.api_broker_queue._reader}

    def run_resolver(self) -> Never:
        """
        The main event loop of the ApiBroker, continuously checking for and processing incoming events from various sources.
        The loop iterates indefinitely until a KeyboardInterrupt is raised, at which point it returns None.

        Returns:
            None: Raises a KeyboardInterrupt exception upon user termination.
        """
        while True:
            pipes: set[Connection] = set()
            if self.router_pipe:
                pipes.add(self.router_pipe)
            if self._workers:
                pipes.update(worker.pipe for worker in self._workers.values())
            pipes.update(self.additional_pipes)
            try:
                active_pipes = wait(pipes, timeout=10)
                for active_pipe in active_pipes:
                    if active_pipe == self.router_pipe:
                        event: Event = self.router_pipe.recv()
                        event._timing.stamp(self.__class__.__name__, "Received")
                        self.logger.debug("Received event: %s", event)
                        self.resolve(event)
                    else:
                        try:
                            event = active_pipe.recv()
                        except EOFError as e:
                            self.logger.error("Communication problems", exc_info=e)
                            continue
                        except OSError as e:
                            self.logger.error("OS problems", exc_info=e)
                            continue
                        else:
                            event._timing.stamp(self.__class__.__name__, "Received")
                            self.resolve_callback(event, active_pipe)

            except KeyboardInterrupt:
                return None

    def resolve_callback(self, event: Event, pipe: Connection) -> None:
        """
        Handles incoming events received from clients or other components, dispatching
        them to the appropriate handling method based on their type.

        Parameters:
            event (Event): The incoming event to be handled.
            pipe (Connection): The communication pipe from which the event was received.

        Returns:
            None
        """
        self.logger.debug("<- Client %s", event)
        if isinstance(event, group.Client):
            if isinstance(event, client.Register):
                self.client_register(event)
            elif isinstance(event, client.Unregister):
                self.client_unregister(event)
            # elif isinstance(event, client.Documentation):
            #     self.documentation(event)
        elif isinstance(event, group.Manage):
            if isinstance(event, manage.Register):
                self.event_register(event)
            elif isinstance(event, manage.Unregister):
                self.event_unregister(event)
        else:
            event_info = self.events.get(event.__class__, None)
            for middleware in self.middlewares_request:
                try:
                    event._timing.stamp(middleware.__class__.__name__, "Started")
                    control = middleware.process_request(event, event_info)
                    event._timing.stamp(middleware.__class__.__name__, "Finished")
                    if not isinstance(control, MiddlewareControl):
                        self.logger.error("Middleware %s suppose to return MiddlewareControl object", middleware.__class__.__name__)
                        event.response.set_status(ResponseStatus.FAILED)
                        event.response.add_directive(InternalServerErrorResponseDirective(
                            f"Middleware  {middleware.__class__.__name__} suppose to return MiddlewareControl object"))
                        self.resolve(event)
                        return

                    if control.action == MiddlewareControl.ActionType.CONTINUE:
                        continue

                    elif control.action == MiddlewareControl.ActionType.STOP:
                        self.logger.info("Processing stopped: %s", middleware.__class__.__name__)
                        break

                    elif control.action == MiddlewareControl.ActionType.REPLACE_EVENT:
                        self.logger.debug("Event replaced by middleware: %s. New event: %s", middleware.__class__.__name__, control.event)
                        event = control.event
                except Exception as e:
                    self.logger.error("Unknown error while processing request in middleware", exc_info=e)
                    event.response.set_status(ResponseStatus.FAILED)
                    event.response.add_directive(InternalServerErrorResponseDirective(
                        f"Unknown error while processing request in middleware: {middleware.__class__.__name__}"))
                    self.resolve(event)
                    return

            self.logger.debug("Router <- API Broker %s", event)
            self.router_queue.put(event)

    def _prepare_resolvers(self) -> None:
        """
        Prepares the resolvers for handling specific types of events, enhancing
        the base implementation with API-specific event handling.

        Returns:
            None
        """
        for event in api_events:
            if "response" not in event.event.__annotations__:
                self._requests[event.event] = self.solve
            else:
                self._responses[event.event] = self.solve
        super()._prepare_resolvers()

    @staticmethod
    def _insert_client_to_trie(node, param_set: dict[str, Any], client: Client) -> None:
        """
        Inserts a client into the trie based on the provided parameter set.

        Parameters:
            node (TrieNode): The current node in the trie.
            param_set (dict[str, Any]): The parameter set defining the subscription path.
            client (Client): The client to be inserted.

        Returns:
            None
        """
        for param, value in param_set.items():
            key = (param, value)
            if key not in node.children:
                node.children[key] = TrieNode(key=key)
            node = node.children[key]
        node.is_end = True
        node.clients.add(client)

    @classmethod
    def _search_trie(cls, node: TrieNode, event: Event, path=None) -> list[Client]:
        """
        Searches the trie to find all clients subscribed to parameters matching the event's attributes.

        Parameters:
            node (TrieNode): The current node in the trie.
            event (Event): The event containing attributes to match.
            path (set): A set to keep track of the current traversal path.

        Returns:
            list[Client]: A list of clients that match the event's parameters.
        """
        if path is None:
            path = set()
        results = []

        if node.is_end:
            results.extend(node.clients)

        for (param, value), child in node.children.items():
            event_fields = {field.name: getattr(event, field.name) for field in fields(event)}
            if param in event_fields and event_fields[param] == value and (param, value) not in path:
                path.add((param, value))
                results.extend(cls._search_trie(child, event, path))
                path.remove((param, value))

        return results

    def _remove_client_from_tries(self, client: Client) -> bool:
        return any([self._remove_client_from_trie(root, client) for root in self.roots.values()])

    @classmethod
    def _remove_client_from_trie(cls, node: TrieNode, client: Client) -> bool:
        """
        Recursively removes the client from the trie. Used when a client unregisters entirely.
        Returns True if the node should be deleted (no clients and no children), False otherwise.

        Parameters:
            node (TrieNode): The current node in the trie.
            client (Client): The client to be removed.

        Returns:
            bool: True if the node can be deleted, False otherwise.
        """
        # Remove the client from the current node if present
        if client in node.clients:
            node.clients.remove(client)

        # Make a list of keys to avoid modifying the dictionary while iterating
        keys = list(node.children.keys())
        for key in keys:
            child = node.children[key]
            # Recursively remove the client from child nodes
            should_delete_child = cls._remove_client_from_trie(child, client)
            if should_delete_child:
                # Delete the child node if it has no clients and no children
                del node.children[key]

        # Determine if the current node should be deleted
        if not node.clients and not node.children:
            return True  # Node can be deleted
        else:
            return False  # Node should remain

    def _remove_client_from_trie_by_param_set(
            self, node: TrieNode, param_items: list[tuple[str, Any]], index: int, client: Client
    ) -> bool:
        """
        Recursive function to remove client from a specific path in trie.
        Returns True if the client was removed, False otherwise.

        Parameters:
            node (TrieNode): The current node in the trie.
            param_items (list[tuple[str, Any]]): The list of parameter items defining the path.
            index (int): The current index in the param_items list.
            client (Client): The client to be removed.

        Returns:
            bool: True if the client was removed, False otherwise.
        """
        if index < len(param_items):
            param, value = param_items[index]
            key = (param, value)
            if key in node.children:
                child = node.children[key]
                client_removed = self._remove_client_from_trie_by_param_set(child, param_items, index + 1, client)
                # After recursion, check if the child node should be deleted
                if client_removed:
                    if not child.clients and not child.children:
                        del node.children[key]
                    return True
                else:
                    return False
            else:
                # The path does not exist; client was not registered for this param_set
                return False
        else:
            # At the leaf node corresponding to param_set
            if client in node.clients:
                node.clients.remove(client)
                return True
            else:
                # Client was not registered for this param_set
                return False

    def _find_root(self, event_name: str) -> TrieNode:
        for event, event_info in self.events.items():
            if event_info.resource == event_name:
                break
        else:
            raise KeyError(event_name)

        return self.roots[event]

    def _subscribe_static_resolvers(self) -> None:
        """
         Registers static resolver methods for specific event types, such as health checks and store get requests.
         """
        self.logger.debug("Registering static resolvers")
        self._requests[HealthCheck] = self.health_check



from multiprocessing.connection import Connection
from dataclasses import dataclass

from edri.config.constant import ApiType


@dataclass(frozen=True, eq=True)
class Client:
    socket: Connection
    type: ApiType
